/* doc.c
 * This file is part of Vara digital painting program
 * Copyright 2021, 2023, 2024 Nandakumar Edamana
 * Distributed under GNU General Public License, version 3
 */ 

#include <assert.h>
#include <stdarg.h>
#include <stdbool.h>

#include "doc.h"

void intvector_construct(intvector *this)
{
	this->UNIT = 8;
	this->count = 0;
	int *_tmp_2 = (int *) calloc((size_t) this->UNIT, sizeof(int));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	this->arr = _tmp_2;
	this->alloccount = this->UNIT;
}

void intvector__resize(intvector *this, int newcount)
{
	int newalloccount = this->alloccount;
	assert((newcount - this->count) == 1 || (newcount - this->count) == -1); /* vector.ngg:22 */
	if(this->alloccount != newcount) {
		if(this->alloccount < newcount) {
			newalloccount = this->alloccount + this->UNIT;
		} else {
			int spare = this->alloccount - newcount;
			if(spare > (2 * this->UNIT)) {
				newalloccount = this->alloccount - this->UNIT;
			}

		}

		assert(newalloccount >= this->count); /* vector.ngg:40 */
		if(newalloccount != this->alloccount) {
			if(newalloccount < 1) {
				newalloccount = 1;
			}

			this->arr = realloc(this->arr, newalloccount * sizeof(this->arr[0]));
			if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

			this->alloccount = newalloccount;
		}

	}

	this->count = newcount;
}

void intvector_append(intvector *this, int newitem)
{
	int newcount = 1 + this->count;
	intvector__resize(this, newcount);
	this->arr[newcount - 1] = newitem;
}

void intvector_clear(intvector *this)
{
	this->count = 0;
	this->alloccount = this->UNIT;
	this->arr = realloc(this->arr, this->UNIT * sizeof(this->arr[0]));
	if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

}

int intvector_get_item(intvector *this, int index)
{
	assert(index < this->count); /* vector.ngg:74 */

	return this->arr[index];
}

int intvector_pop(intvector *this)
{
	assert(this->count > 0); /* vector.ngg:80 */
	int r = intvector_get_item(this, this->count - 1);
	this->count = this->count - 1;

	return r;
}

void intvector_set_item(intvector *this, int index, int itm)
{
	assert(index < this->count); /* vector.ngg:93 */
	this->arr[index] = itm;
}

int intvector_get_count(intvector *this)
{
	return this->count;
}

_Bool intvector_is_empty(intvector *this)
{
	return 0 == this->count;
}

void intvector_destruct(intvector *this)
{
	if(this->arr) {
		free(this->arr);
	}

}

struct Pixel PIXEL_CHECKER_DARK = (struct Pixel) { {0.5, 0.5, 0.5, 1} };
struct Pixel PIXEL_CHECKER_LIGHT = (struct Pixel) { {0.8, 0.8, 0.8, 1} };
struct Pixel PIXEL_SCRATCH_DARK = (struct Pixel) { {0.1, 0.1, 0.1, 1} };
struct Pixel PIXEL_SCRATCH_LIGHT = (struct Pixel) { {0.8, 0.8, 0, 1} };
void layer_owning_vector_construct(LayerOwningVector *this)
{
	this->UNIT = 8;
	this->alloccount = 8;
	this->count = 0;
	Layer * *_tmp_2 = (Layer * *) calloc((size_t) this->UNIT, sizeof(Layer));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	this->arr = _tmp_2;
	this->alloccount = this->UNIT;
}

void layer_owning_vector_destruct(LayerOwningVector *this)
{
	layer_owning_vector_delete_all(this);
	if(this->arr) {
		free(this->arr);
	}

}

void layer_owning_vector__resize(LayerOwningVector *this, int newcount)
{
	int newalloccount = this->alloccount;
	assert((newcount - this->count) == 1 || (newcount - this->count) == -1); /* owningvector.ngg:30 */
	if(this->alloccount != newcount) {
		if(this->alloccount < newcount) {
			newalloccount = this->alloccount + this->UNIT;
		} else if(this->alloccount > (newcount + this->UNIT)) {
			assert((this->alloccount - this->UNIT) >= 0); /* owningvector.ngg:41 */
			newalloccount = this->alloccount - this->UNIT;
		}

		assert(newalloccount >= this->count); /* owningvector.ngg:45 */
		if(newalloccount != this->alloccount) {
			if(newalloccount < 1) {
				newalloccount = 1;
			}

			this->arr = realloc(this->arr, newalloccount * sizeof(this->arr[0]));
			if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

			this->alloccount = newalloccount;
		}

	}

	this->count = newcount;
}

void layer_owning_vector_append(LayerOwningVector *this, Layer *newitem)
{
	int newcount = 1 + this->count;
	layer_owning_vector__resize(this, newcount);
	this->arr[newcount - 1] = newitem;
}

void layer_owning_vector_clear(LayerOwningVector *this)
{
	layer_owning_vector_delete_all(this);
	this->count = 0;
	this->arr = realloc(this->arr, this->UNIT * sizeof(this->arr[0]));
	if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

	this->alloccount = this->UNIT;
}

void layer_owning_vector_delete_all(LayerOwningVector *this)
{
	int i;
	for(i = 0; i < this->count; i += 1) {
		if(this->arr[i]) {
			layer_destruct(this->arr[i]);
			free(this->arr[i]);
		}

	}

}

Layer * layer_owning_vector_get_item(LayerOwningVector *this, int index)
{
	assert(index < this->count); /* owningvector.ngg:91 */

	return this->arr[index];
}

Layer * layer_owning_vector_pop(LayerOwningVector *this)
{
	assert(this->count > 0); /* owningvector.ngg:97 */
	Layer *r = layer_owning_vector_get_item(this, this->count - 1);
	this->count = this->count - 1;

	return r;
}

void layer_owning_vector_set_item(LayerOwningVector *this, int index, Layer *itm)
{
	assert(index < this->count); /* owningvector.ngg:110 */
	if(this->arr[index]) {
		layer_destruct(this->arr[index]);
	}

	this->arr[index] = itm;
}

int layer_owning_vector_get_count(LayerOwningVector *this)
{
	return this->count;
}

_Bool layer_owning_vector_is_empty(LayerOwningVector *this)
{
	return 0 == this->count;
}

void history_point_construct(HistoryPoint *this, int layeridx, Image *img)
{
	this->img = image_Clone(img);
	this->layeridx = layeridx;
}

Image * history_point_detach_img(HistoryPoint *this)
{
	assert(this->img);
	Image *ret = this->img;
	this->img = NULL;

	return ret;
}

void history_point_destruct(HistoryPoint *this)
{
	if(this->img) {
		image_destruct((Image *) this->img);
		free(this->img);
	}

}

void history_point_stack_construct(HistoryPointStack *this)
{
	this->count = 0;
	HistoryPoint * *_tmp_2 = (HistoryPoint * *) calloc((size_t) 8, sizeof(HistoryPoint));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	this->arr = _tmp_2;
	this->alloccount = 8;
}

_Bool history_point_stack_is_empty(HistoryPointStack *this)
{
	return 0 == this->count;
}

void history_point_stack_push(HistoryPointStack *this, HistoryPoint *newitem)
{
	history_point_stack__resize(this, this->count + 1);
	this->arr[this->count - 1] = newitem;
}

void history_point_stack_push_multiple(HistoryPointStack *this, int n, HistoryPoint * *newitems)
{
	int i;
	int newcount = this->count + n;
	if(this->alloccount < newcount) {
		this->arr = realloc(this->arr, newcount * sizeof(this->arr[0]));
		if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

		this->alloccount = newcount;
	}

	for(i = 0; i < n; i += 1) {
		this->arr[this->count + i] = newitems[i];
	}

	this->count = newcount;
}

HistoryPoint * history_point_stack_peek(HistoryPointStack *this)
{
	assert(this->count > 0); /* stack.ngg:42 */

	return this->arr[this->count - 1];
}

HistoryPoint * history_point_stack_pop(HistoryPointStack *this)
{
	assert(this->count > 0); /* stack.ngg:48 */
	HistoryPoint *retitem = this->arr[this->count - 1];
	history_point_stack__resize(this, this->count - 1);

	return retitem;
}

HistoryPoint * history_point_stack_pop_bottom(HistoryPointStack *this)
{
	assert(this->count > 0); /* stack.ngg:61 */
	HistoryPoint *retval = this->arr[0];
	this->count -= 1;
	memmove(this->arr, this->arr + 1, this->count * sizeof(this->arr[0]));

	return retval;
}

void history_point_stack_replace_top(HistoryPointStack *this, HistoryPoint *newitem)
{
	assert(this->count > 0); /* stack.ngg:73 */
	this->arr[this->count - 1] = newitem;
}

void history_point_stack__resize(HistoryPointStack *this, int newcount)
{
	int UNIT = 8;
	int newalloccount = this->alloccount;
	if(this->alloccount != newcount) {
		if(this->alloccount < newcount) {
			newalloccount = this->alloccount + UNIT;
		} else {
			int spare = this->alloccount - newcount;
			if(spare > (2 * UNIT)) {
				newalloccount = this->alloccount - UNIT;
			}

		}

		assert(newalloccount >= this->count); /* stack.ngg:104 */
		if(newalloccount != this->alloccount) {
			if(newalloccount < 1) {
				newalloccount = 1;
			}

			this->arr = realloc(this->arr, newalloccount * sizeof(this->arr[0]));
			if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

			this->alloccount = newalloccount;
		}

	}

	this->count = newcount;
}

void history_point_stack_destruct(HistoryPointStack *this)
{
	if(this->arr) {
		free(this->arr);
	}

}

void image_document_construct(ImageDocument *this, int width, int height, struct DrawingParams drawparams)
{
	HistoryPointStack *_tmp_2 = (HistoryPointStack *) malloc(sizeof(HistoryPointStack));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	history_point_stack_construct(_tmp_2);
	this->redostack = _tmp_2;
	HistoryPointStack *_tmp_4 = (HistoryPointStack *) malloc(sizeof(HistoryPointStack));
	if(_tmp_4 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	history_point_stack_construct(_tmp_4);
	this->undostack = _tmp_4;
	intvector *_tmp_5 = (intvector *) malloc(sizeof(intvector));
	if(_tmp_5 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	intvector_construct(_tmp_5);
	this->layerstack = _tmp_5;
	LayerOwningVector *_tmp_6 = (LayerOwningVector *) malloc(sizeof(LayerOwningVector));
	if(_tmp_6 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	layer_owning_vector_construct(_tmp_6);
	this->layers = _tmp_6;
	this->history_changed_userdata = NULL;
	this->history_changed_cbk = NULL;
	this->prescratchbak = NULL;
	this->autocenter = false;
	this->oldpreviewzoom = 0.0;
	this->oldpreviewh = 0;
	this->oldprevieww = 0;
	this->oldpreviewy = 0;
	this->oldpreviewx = 0;
	this->vpy = 0;
	this->vpx = 0;
	this->scratchmode = false;
	this->oldy = 0.0;
	this->oldx = 0.0;
	this->pen_is_down = false;
	this->preview_is_all_dirty = false;
	this->scratchborder = NULL;
	this->preview = NULL;
	this->dirtyrect = vara_rect_default();
	assert(width > 0); /* doc.ngg:114 */
	assert(height > 0); /* doc.ngg:115 */
	this->bgcolor = LINEAR_PIXEL_WHITE;
	Image *_tmp_7 = (Image *) malloc(sizeof(Image));
	if(_tmp_7 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(_tmp_7, width, height, 255);
	this->checkpattern = _tmp_7;
	image_fill_checkpattern(this->checkpattern);
	this->drawparams = drawparams;
	this->stroke_smoothness = 0.1;
	Layer *_tmp_8 = (Layer *) malloc(sizeof(Layer));
	if(_tmp_8 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	layer_construct(_tmp_8, width, height, LINEAR_PIXEL_WHITE);
	layer_owning_vector_append(this->layers, _tmp_8);
	Layer *_tmp_9 = (Layer *) malloc(sizeof(Layer));
	if(_tmp_9 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	layer_construct(_tmp_9, width, height, LINEAR_PIXEL_TRANSP);
	layer_owning_vector_append(this->layers, _tmp_9);
	intvector_append(this->layerstack, 0);
	intvector_append(this->layerstack, 1);
	this->curlayeridxidx = 1;
	this->subpixel_when_zoomed = true;
	this->vpzoom = 1;
	Image *_tmp_10 = (Image *) malloc(sizeof(Image));
	if(_tmp_10 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(_tmp_10, width, height, 255);
	this->composite = _tmp_10;
	image_document_mark_all_dirty(this);
}

void image_document_destruct(ImageDocument *this)
{
	image_document_clear_undostack(this);
	image_document_clear_redostack(this);
	if(this->checkpattern) {
		image_destruct(this->checkpattern);
		free(this->checkpattern);
	}

	if(this->layers) {
		layer_owning_vector_destruct(this->layers);
		free(this->layers);
	}

	if(this->layerstack) {
		intvector_destruct(this->layerstack);
		free(this->layerstack);
	}

	if(this->composite) {
		image_destruct(this->composite);
		free(this->composite);
	}

	if(this->preview) {
		image_destruct((Image *) this->preview);
		free(this->preview);
	}

	if(this->scratchborder) {
		image_destruct((Image *) this->scratchborder);
		free(this->scratchborder);
	}

	if(this->undostack) {
		history_point_stack_destruct(this->undostack);
		free(this->undostack);
	}

	if(this->redostack) {
		history_point_stack_destruct(this->redostack);
		free(this->redostack);
	}

	if(this->prescratchbak) {
		history_point_destruct((HistoryPoint *) this->prescratchbak);
		free(this->prescratchbak);
	}

}

void image_document_apply_snapshot(ImageDocument *this, HistoryPoint *hp)
{
	layer_replace_image(layer_owning_vector_get_item(this->layers, hp->layeridx), history_point_detach_img(hp));
	image_document_mark_all_dirty(this);
}

_Bool image_document_can_redo(ImageDocument *this)
{
	return (!this->scratchmode) && (!history_point_stack_is_empty(this->redostack));
}

_Bool image_document_can_switch_layer(ImageDocument *this)
{
	return !this->scratchmode;
}

_Bool image_document_can_undo(ImageDocument *this)
{
	return (!this->scratchmode) && (!history_point_stack_is_empty(this->undostack));
}

void image_document_clear_layer(ImageDocument *this)
{
	image_document_fill_layer(this, LINEAR_PIXEL_TRANSP);
}

void image_document_clear_redostack(ImageDocument *this)
{
	while(!history_point_stack_is_empty(this->redostack)) {
		HistoryPoint *todel = history_point_stack_pop(this->redostack);
		if(todel) {
			history_point_destruct(todel);
			free(todel);
		}

	}

}

void image_document_clear_undostack(ImageDocument *this)
{
	while(!history_point_stack_is_empty(this->undostack)) {
		HistoryPoint *todel = history_point_stack_pop(this->undostack);
		if(todel) {
			history_point_destruct(todel);
			free(todel);
		}

	}

}

void image_document_composite_layerstack(ImageDocument *this, Image *target, _Bool all_dirty, struct VaraRect dirtyrect)
{
	int _ngg_tmp_3;
	int idx;
	for(_ngg_tmp_3 = 0; _ngg_tmp_3 < intvector_get_count(this->layerstack); _ngg_tmp_3 += 1) {
		idx = intvector_get_item(this->layerstack, _ngg_tmp_3);
		Layer *layer = layer_owning_vector_get_item(this->layers, idx);
		if(layer->visible) {
			if(all_dirty) {
				image_overlay_image_with_alpha(target, layer->img, layer->alpha);
			} else {
				image_overlay_image_region_with_alpha(target, layer->img, layer->alpha, dirtyrect);
			}

		}

	}

}

Image * image_document_create_export_composite(ImageDocument *this, _Bool fillbg)
{
	Image *ecomposite = (Image *) malloc(sizeof(Image));
	if(ecomposite == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(ecomposite, this->composite->width, this->composite->height, 255);
	if(fillbg) {
		image_fill(ecomposite, this->bgcolor);
	}

	image_document_composite_layerstack(this, ecomposite, true, vara_rect_default());

	return ecomposite;
}

HistoryPoint * image_document_create_history_point(ImageDocument *this)
{
	HistoryPoint *_tmp_2 = (HistoryPoint *) malloc(sizeof(HistoryPoint));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	history_point_construct(_tmp_2, image_document_get_curlayeridx(this), image_document_get_curlayer(this)->img);

	return _tmp_2;
}

void image_document_create_redo_snapshot(ImageDocument *this)
{
	history_point_stack_push(this->redostack, image_document_create_history_point(this));
	if(this->history_changed_cbk) {
		void (*cbknn)(void *userdata) = (void (*)(void *userdata)) this->history_changed_cbk;
		cbknn(this->history_changed_userdata);
	}

}

void image_document_create_undo_snapshot(ImageDocument *this)
{
	if(this->scratchmode) {
		return ;
	}

	image_document_clear_redostack(this);
	if(this->undostack->count > STACKLIMIT) {
		HistoryPoint *todel = history_point_stack_pop_bottom(this->undostack);
		assert(this->undostack->count == STACKLIMIT); /* doc.ngg:234 */
		if(todel) {
			history_point_destruct(todel);
			free(todel);
		}

	}

	history_point_stack_push(this->undostack, image_document_create_history_point(this));
	if(this->history_changed_cbk) {
		void (*cbknn)(void *userdata) = (void (*)(void *userdata)) this->history_changed_cbk;
		cbknn(this->history_changed_userdata);
	}

}

Layer * image_document_get_curlayer(ImageDocument *this)
{
	return layer_owning_vector_get_item(this->layers, image_document_get_curlayeridx(this));
}

int image_document_get_curlayeridx(ImageDocument *this)
{
	return intvector_get_item(this->layerstack, image_document_get_curlayeridxidx(this));
}

int image_document_get_curlayeridxidx(ImageDocument *this)
{
	assert(this->curlayeridxidx >= 0 && this->curlayeridxidx <= (intvector_get_count(this->layerstack) - 1)); /* doc.ngg:253 */

	return this->curlayeridxidx;
}

Layer * image_document_get_bglayer(ImageDocument *this)
{
	return layer_owning_vector_get_item(this->layers, 0);
}

Layer * image_document_get_fglayer(ImageDocument *this)
{
	return layer_owning_vector_get_item(this->layers, 1);
}

void image_document_dip_eyedropper(ImageDocument *this, int x, int y)
{
	assert(this->drawparams.active_tool == TOOL_EYEDROPPER); /* doc.ngg:262 */
	this->drawparams.fgcolor = image_document_get_color_at(this, x, y);
}

enum VaraFileIOError image_document_export_document(ImageDocument *this, const char * filename, _Bool fillbg)
{
	Image *ecomposite = image_document_create_export_composite(this, fillbg);
	enum VaraFileIOError _tmp_2 = write_image_to_file(ecomposite, filename);
	if(ecomposite) {
		image_destruct(ecomposite);
		free(ecomposite);
	}


	return _tmp_2;
}

enum VaraFileIOError image_document_export_layer(ImageDocument *this, const char * filename, _Bool fillbg)
{
	Layer *curlayer = image_document_get_curlayer(this);
	Image *comp = (Image *) malloc(sizeof(Image));
	if(comp == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(comp, curlayer->img->width, curlayer->img->height, curlayer->img->depth);
	if(fillbg) {
		image_fill(comp, this->bgcolor);
	}

	image_overlay_image_with_alpha(comp, curlayer->img, curlayer->alpha);
	enum VaraFileIOError _tmp_2 = write_image_to_file(comp, filename);
	if(comp) {
		image_destruct(comp);
		free(comp);
	}


	return _tmp_2;
}

void image_document_fill_layer(ImageDocument *this, struct LinearPixel color)
{
	image_document_create_undo_snapshot(this);
	image_fill(image_document_get_curlayer(this)->img, color);
	image_document_mark_all_dirty(this);
}

void image_document_fill_layer_with_bgcolor(ImageDocument *this)
{
	image_document_fill_layer(this, image_document_get_curlayer(this)->default_bgcolor);
}

void image_document_fill_layer_with_fgcolor(ImageDocument *this)
{
	image_document_fill_layer(this, this->drawparams.fgcolor);
}

void image_document_floodfill(ImageDocument *this, int x, int y)
{
	assert(this->drawparams.active_tool == TOOL_FLOODFILL); /* doc.ngg:304 */
	double tx = image_document_translate_x(this, x);
	double ty = image_document_translate_y(this, y);
	if(image_is_point_outside(image_document_get_curlayer(this)->img, (int) tx, (int) ty)) {
		return ;
	}

	image_document_create_undo_snapshot(this);
	struct VaraRect rect = nan_img_floodfill(image_document_get_curlayer(this)->img, (int) tx, (int) ty, this->drawparams);
	image_document_add_dirty_region(this, rect);
}

struct LinearPixel image_document_get_color_at(ImageDocument *this, int x, int y)
{
	double tx = image_document_translate_x(this, x);
	double ty = image_document_translate_y(this, y);
	if((tx >= 0) && (tx < image_document_get_curlayer(this)->img->width) && (ty >= 0) && (ty < image_document_get_curlayer(this)->img->height)) {
		struct LinearPixel px = nan_img_get_linear_pixel(image_document_get_curlayer(this)->img, (int) tx, (int) ty);
		px.rgb[3] = 1;

		return px;
	}


	return LINEAR_PIXEL_BLACK;
}

int image_document_get_height(ImageDocument *this)
{
	return image_document_get_fglayer(this)->img->height;
}

int image_document_get_width(ImageDocument *this)
{
	return image_document_get_fglayer(this)->img->width;
}

_Bool image_document_is_curlayer_bg(ImageDocument *this)
{
	return image_document_get_curlayeridxidx(this) == 0;
}

_Bool image_document_is_curlayer_fg(ImageDocument *this)
{
	return image_document_get_curlayeridxidx(this) == 1;
}

void image_document_add_dirty_region(ImageDocument *this, struct VaraRect rect)
{
	if(region_is_zero(this->dirtyrect)) {
		this->dirtyrect = rect;

		return ;
	}

	this->dirtyrect = union_regions(this->dirtyrect, rect);
}

_Bool image_document_is_all_dirty(ImageDocument *this)
{
	return (this->dirtyrect.x == 0) && (this->dirtyrect.y == 0) && (this->dirtyrect.w == this->composite->width) && (this->dirtyrect.h == this->composite->height);
}

_Bool image_document_is_dirty(ImageDocument *this)
{
	return !region_is_zero(this->dirtyrect);
}

void image_document_mark_all_dirty(ImageDocument *this)
{
	this->dirtyrect.x = 0;
	this->dirtyrect.y = 0;
	this->dirtyrect.w = this->composite->width;
	this->dirtyrect.h = this->composite->height;
}

void image_document_mark_not_dirty(ImageDocument *this)
{
	this->dirtyrect.w = 0;
	this->dirtyrect.h = 0;
}

void image_document_pen_down(ImageDocument *this, int x, int y, double pressure)
{
	image_document_create_undo_snapshot(this);
	this->pen_is_down = true;
	this->oldx = x;
	this->oldy = y;
	image_document_pen_stroke(this, x, y, pressure);
}

void image_document_pen_up(ImageDocument *this, int x, int y, double pressure)
{
	image_document_pen_stroke(this, x, y, pressure);
	this->pen_is_down = false;
}

void image_document_pen_stroke(ImageDocument *this, int ix, int iy, double pressure)
{
	image_document_update_pressure(this, pressure);
	double x = ix;
	double y = iy;
	if(this->stroke_smoothness > 0) {
		double dist = this->stroke_smoothness * 20;
		x = this->oldx + ((x - this->oldx) / dist);
		y = this->oldy + ((y - this->oldy) / dist);
	}

	double tx1 = image_document_translate_x(this, this->oldx);
	double ty1 = image_document_translate_y(this, this->oldy);
	double tx2 = image_document_translate_x(this, x);
	double ty2 = image_document_translate_y(this, y);
	if((!this->subpixel_when_zoomed) && (this->stroke_smoothness <= 0)) {
		tx1 = (int) tx1;
		ty1 = (int) ty1;
		tx2 = (int) tx2;
		ty2 = (int) ty2;
	}

	struct VaraRect rect = image_line(image_document_get_curlayer(this)->img, tx1, ty1, tx2, ty2, this->drawparams);
	image_document_add_dirty_region(this, rect);
	this->oldx = x;
	this->oldy = y;
}

void image_document_switch_layer(ImageDocument *this, int stackidx)
{
	if(!image_document_can_switch_layer(this)) {
		vara_log_error("attempt to switch layer when it is disabled");

		return ;
	}

	assert(stackidx >= 0 && stackidx <= (intvector_get_count(this->layerstack) - 1)); /* doc.ngg:441 */
	this->curlayeridxidx = stackidx;
}

void image_document_switch_layer_bg(ImageDocument *this)
{
	image_document_switch_layer(this, 0);
}

void image_document_switch_layer_fg(ImageDocument *this)
{
	image_document_switch_layer(this, 1);
}

double image_document_translate_x(ImageDocument *this, double x)
{
	assert(this->vpzoom > 0); /* doc.ngg:454 */

	return (x - this->vpx) / this->vpzoom;
}

double image_document_translate_y(ImageDocument *this, double y)
{
	assert(this->vpzoom > 0); /* doc.ngg:459 */

	return (y - this->vpy) / this->vpzoom;
}

void image_document_update_pressure(ImageDocument *this, double pressure)
{
	this->drawparams.oldpressure = this->drawparams.pressure;
	this->drawparams.pressure = pressure;
}

void image_document_undo(ImageDocument *this)
{
	if(!image_document_can_undo(this)) {
		vara_log_error("undo issued when it is disabled");

		return ;
	}

	history_point_stack_push(this->redostack, image_document_create_history_point(this));
	HistoryPoint *_ngg_tmp_1 = history_point_stack_pop(this->undostack);
	image_document_apply_snapshot(this, _ngg_tmp_1);
	if(this->history_changed_cbk) {
		void (*cbknn)(void *userdata) = (void (*)(void *userdata)) this->history_changed_cbk;
		cbknn(this->history_changed_userdata);
	}

	if(_ngg_tmp_1) {
		history_point_destruct(_ngg_tmp_1);
		free(_ngg_tmp_1);
	}

}

void image_document_viewport_align_center(ImageDocument *this)
{
	assert(this->preview);
	Image *preview = this->preview;
	this->vpx = (int) ((preview->width / 2.0) - ((this->vpzoom * this->composite->width) / 2.0));
	this->vpy = (int) ((preview->height / 2.0) - ((this->vpzoom * this->composite->height) / 2.0));
	this->autocenter = true;
}

void image_document_viewport_set_offsetx(ImageDocument *this, int x)
{
	this->vpx = x;
	this->autocenter = false;
}

void image_document_viewport_set_offsety(ImageDocument *this, int y)
{
	this->vpy = y;
	this->autocenter = false;
}

void image_document_viewport_set_size(ImageDocument *this, int w, int h)
{
	if(this->preview) {
		Image *previewnn = (Image *) this->preview;
		if((w == previewnn->width) && (h == previewnn->height)) {
			return ;
		}

	}

	if(this->preview) {
		image_destruct((Image *) this->preview);
		free(this->preview);
		this->preview = NULL;
	}

	Image *_tmp_2 = (Image *) malloc(sizeof(Image));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(_tmp_2, w, h, 255);
	this->preview = _tmp_2;
	Image *sbimg = (Image *) malloc(sizeof(Image));
	if(sbimg == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(sbimg, w, SCRATCHBORDER, 255);
	image_fill_scratchpattern(sbimg);
	if(this->scratchborder) {
		Image *scratchbordernn = (Image *) this->scratchborder;
		if(this->scratchborder) {
			image_destruct((Image *) this->scratchborder);
			free(this->scratchborder);
			this->scratchborder = NULL;
		}

	}

	this->scratchborder = sbimg;
	if(this->autocenter) {
		image_document_viewport_align_center(this);
	}

}

void image_document_viewport_zoom_at(ImageDocument *this, int x, int y, double zoom)
{
	image_document_viewport_zoom_at_internal(this, x, y, zoom);
	this->autocenter = false;
}

void image_document_viewport_zoom_at_internal(ImageDocument *this, int x, int y, double zoom)
{
	double srcx = image_document_translate_x(this, x);
	double srcy = image_document_translate_y(this, y);
	this->vpzoom = zoom;
	this->vpx = (int) (x - (this->vpzoom * srcx));
	this->vpy = (int) (y - (this->vpzoom * srcy));
}

void image_document_viewport_zoom_at_center(ImageDocument *this, double zoom)
{
	assert(this->preview);
	Image *preview = this->preview;
	image_document_viewport_zoom_at_internal(this, preview->width / 2, preview->height / 2, zoom);
}

_ngg_tuple_ImageDocument__update_composite image_document_update_composite(ImageDocument *this)
{
	if(!image_document_is_dirty(this)) {
		struct VaraRect defrect = vara_rect_default();

		return (_ngg_tuple_ImageDocument__update_composite){false, defrect};
	}

	image_copy_image_region(this->composite, this->checkpattern, this->dirtyrect);
	image_document_composite_layerstack(this, this->composite, image_document_is_all_dirty(this), this->dirtyrect);
	if(DEBUG_COMPOSITE_DIRTYRECT) {
		print_region(this->dirtyrect, "composite dirtyrect");
		image_draw_debug_region(this->composite, this->dirtyrect);
	}

	struct VaraRect prvdirtyrect = this->dirtyrect;
	image_document_mark_not_dirty(this);

	return (_ngg_tuple_ImageDocument__update_composite){true, prvdirtyrect};
}

struct VaraRect image_document_update_preview(ImageDocument *this)
{
	assert(this->preview);
	Image *preview = this->preview;
	_ngg_tuple_ImageDocument__update_composite _ngg_tmp_0 = image_document_update_composite(this);
	struct VaraRect compdirtyrect = _ngg_tmp_0.m1;
	_Bool isdirty = _ngg_tmp_0.m0;
	if(isdirty) {
		if(REPREVIEW_DIRTY_ONLY && SCALING_AA_ENABLE) {
			double border = this->vpzoom * SCALING_AA_RADIUS;
			compdirtyrect.x -= border;
			compdirtyrect.y -= border;
			compdirtyrect.w += (2 * border);
			compdirtyrect.h += (2 * border);
			compdirtyrect = image_clip_region(this->composite, compdirtyrect);
		}

	}

	_Bool geometry_changed = ((((this->oldpreviewx != this->vpx) || (this->oldpreviewy != this->vpy)) || (this->oldprevieww != preview->width)) || (this->oldpreviewh != preview->height)) || (this->oldpreviewzoom != this->vpzoom);
	_Bool redraw_all = (!REPREVIEW_DIRTY_ONLY) || geometry_changed;
	if(this->preview_is_all_dirty) {
		redraw_all = true;
		this->preview_is_all_dirty = false;
	}

	if(redraw_all) {
		image_fill(preview, srgbpixel2linear(PIXEL_VPBG));
	}

	struct VaraRect prevdirtyrect = vara_rect_default();
	if(this->vpzoom == 1) {
		if(redraw_all) {
			prevdirtyrect = image_copy_image_with_destoffset(preview, this->composite, this->vpx, this->vpy);
		} else if(isdirty) {
			prevdirtyrect = image_copy_image_region_with_destoffset(preview, this->composite, compdirtyrect, this->vpx, this->vpy);
		}

	} else {
		if(redraw_all) {
			prevdirtyrect = image_copy_image_with_scale(preview, this->composite, this->vpx, this->vpy, this->vpzoom);
		} else if(isdirty) {
			prevdirtyrect = image_copy_image_region_with_scale(preview, this->composite, compdirtyrect, this->vpx, this->vpy, this->vpzoom);
		}

	}

	this->oldpreviewx = this->vpx;
	this->oldpreviewy = this->vpy;
	this->oldprevieww = preview->width;
	this->oldpreviewh = preview->height;
	this->oldpreviewzoom = this->vpzoom;
	if(this->scratchmode) {
		assert(this->scratchborder);
		Image *sbimg = this->scratchborder;
		image_copy_image_with_destoffset(preview, sbimg, 0, 0);
		image_copy_image_with_destoffset(preview, sbimg, 0, preview->height - SCRATCHBORDER);

		return image_get_region(preview);
	}

	if(redraw_all) {
		return image_get_region(preview);
	}


	return prevdirtyrect;
}

void image_document_redo(ImageDocument *this)
{
	if(!image_document_can_redo(this)) {
		vara_log_error("redo issued when it is disabled");

		return ;
	}

	history_point_stack_push(this->undostack, image_document_create_history_point(this));
	HistoryPoint *_ngg_tmp_2 = history_point_stack_pop(this->redostack);
	image_document_apply_snapshot(this, _ngg_tmp_2);
	if(this->history_changed_cbk) {
		void (*cbknn)(void *userdata) = (void (*)(void *userdata)) this->history_changed_cbk;
		cbknn(this->history_changed_userdata);
	}

	if(_ngg_tmp_2) {
		history_point_destruct(_ngg_tmp_2);
		free(_ngg_tmp_2);
	}

}

void image_document_set_bgvisible(ImageDocument *this, _Bool visible)
{
	layer_set_visible(image_document_get_bglayer(this), visible);
	image_document_mark_all_dirty(this);
}

void image_document_set_fgvisible(ImageDocument *this, _Bool visible)
{
	layer_set_visible(image_document_get_fglayer(this), visible);
	image_document_mark_all_dirty(this);
}

void image_document_set_history_changed_callback(ImageDocument *this, void (*cbk)(void *userdata), void *userdata)
{
	this->history_changed_cbk = cbk;
	this->history_changed_userdata = userdata;
}

void image_document_toggle_scratch_mode(ImageDocument *this)
{
	if(this->scratchmode) {
		assert(this->prescratchbak);
		image_document_apply_snapshot(this, this->prescratchbak);
		if(this->prescratchbak) {
			history_point_destruct((HistoryPoint *) this->prescratchbak);
			free(this->prescratchbak);
			this->prescratchbak = NULL;
		}

	} else {
		assert(!this->prescratchbak); /* doc.ngg:699 */
		this->prescratchbak = image_document_create_history_point(this);
	}

	this->scratchmode = !this->scratchmode;
	this->preview_is_all_dirty = true;
}

void image_fill_checkpattern(Image *img)
{
	int y;
	struct LinearPixel lightpx = srgbpixel2linear(PIXEL_CHECKER_LIGHT);
	struct LinearPixel darkpx = srgbpixel2linear(PIXEL_CHECKER_DARK);
	for(y = 0; y < img->height; y += 1) {
		int x;
		for(x = 0; x < img->width; x += 1) {
			_Bool islight = 0 != (((x >> 3) & 1) ^ ((y >> 3) & 1));
			nan_img_set_linear_pixel(img, x, y, (islight ? lightpx : darkpx));
		}

	}

}

void image_fill_scratchpattern(Image *img)
{
	int y;
	struct LinearPixel lightpx = srgbpixel2linear(PIXEL_SCRATCH_LIGHT);
	struct LinearPixel darkpx = srgbpixel2linear(PIXEL_SCRATCH_DARK);
	for(y = 0; y < imin2(SCRATCHBORDER, img->height); y += 1) {
		int x;
		for(x = 0; x < img->width; x += 1) {
			int islight = ((x + y) / SCRATCHBORDER) & 1;
			nan_img_set_linear_pixel(img, x, y, (islight ? lightpx : darkpx));
		}

	}

}

enum VaraFileIOError write_image_to_file(Image *img, const char * fname)
{
	return write_image_to_png(img, fname);
}

void vara_log_error(const char * msg)
{
	fprintf(stderr, "vara: error: %s\n", msg);
}

_ngg_tuple_ImageDocument__update_composite _ngg_tuple__image_document__update_composite_default()
{
	_ngg_tuple_ImageDocument__update_composite s;
	s.m0 = false;
	s.m1 = vara_rect_default();

	return s;
}

